@@ -18,11 +18,14 @@
 import os.path
 from hashlib import md5
 import re
+import datetime
 import gettext
 from gettext import gettext as _
+from typing import Any
+from collections.abc import Callable
 import gi
-gi.require_version('Gtk', '3.0')
-from gi.repository import Gtk
+gi.require_version('Gtk', '4.0')
+from gi.repository import Gtk, Gio # NOQA: E402
 gettext.textdomain('gTranscribe')
 
 # pylint: disable=invalid-name
@@ -35,58 +38,60 @@ class duration:
 
     __slots__ = '_hour', '_minute', '_second', '_microsecond'
 
-    def __new__(cls, hour=0, minute=0, second=0, microsecond=0):
+    def __new__(cls, hour: int = 0, minute: int = 0, second: int = 0, microsecond: int = 0) -> 'duration':
         """Constructor.
         Arguments:
         hour, minute, second, microsecond (default to zero)
         """
         self = object.__new__(cls)
-        self._hour = hour
-        self._minute = minute
-        self._second = second
-        self._microsecond = microsecond
+        self._hour = hour  # type: ignore[attr-defined]
+        self._minute = minute  # type: ignore[attr-defined]
+        self._second = second  # type: ignore[attr-defined]
+        self._microsecond = microsecond  # type: ignore[attr-defined]
         return self
 
     @property
-    def hour(self):
+    def hour(self) -> int:
         """hour"""
-        return self._hour
+        return self._hour  # type: ignore[attr-defined,no-any-return]
 
     @property
-    def minute(self):
+    def minute(self) -> int:
         """minute (0-59)"""
-        return self._minute
+        return self._minute  # type: ignore[attr-defined,no-any-return]
 
     @property
-    def second(self):
+    def second(self) -> int:
         """second (0-59)"""
-        return self._second
+        return self._second  # type: ignore[attr-defined,no-any-return]
 
     @property
-    def microsecond(self):
+    def microsecond(self) -> int:
         """microsecond (0-999999)"""
-        return self._microsecond
+        return self._microsecond  # type: ignore[attr-defined,no-any-return]
 
-    def strftime(self, fmt_string):
+    def strftime(self, fmt_string: str) -> str:
         """
         Probably not what you think it is :-(.
         It just implements what's needed for gTranscribe.
         """
-        txt = fmt_string.replace("%H", str(self._hour))
-        txt = txt.replace("%M", str(self._minute))
-        txt = txt.replace("%S", str(self._second))
-        txt = txt.replace("%f", str(self._microsecond))
+        txt = fmt_string.replace("%H", f"{self._hour:02d}")  # type: ignore[attr-defined]
+        txt = txt.replace("%M", f"{self._minute:02d}")  # type: ignore[attr-defined]
+        txt = txt.replace("%S", f"{self._second:02d}")  # type: ignore[attr-defined]
+        txt = txt.replace("%f", str(self._microsecond))  # type: ignore[attr-defined]
         return txt
 
 
-def trim(timestring, digits=1):
+def trim(timestring: str, digits: int = 1) -> str:
     """Trim a time string to contain only a given number of digits."""
-    pos = re.search("[.-]", timestring).start()
-    if pos > -1:
-        return timestring[:pos + 1 + digits]
+    match = re.search("[.-]", timestring)
+    if match:
+        pos = match.start()
+        if pos > -1:
+            return timestring[:pos + 1 + digits]
     return timestring
 
-def ns_to_time(ns):
+def ns_to_time(ns: int) -> duration:
     """
     Convert nanoseconds to a duration object.
 
@@ -106,7 +111,7 @@ def ns_to_time(ns):
     return duration(h, m, s, ms)
 
 
-def time_to_ns(time):
+def time_to_ns(time: duration | datetime.time) -> int:
     """
     Convert a datetime.time or duration object to nanoseconds.
 
@@ -124,53 +129,34 @@ def time_to_ns(time):
     return sum(values)
 
 
-def get_open_filename(self, title, filter_name, filter_mime):
-    """Display a file open dialog and returns the filename."""
-    filename = None
-    chooser = Gtk.FileChooserDialog(title, self.window,
-                                    Gtk.FileChooserAction.OPEN,
-                                    (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
-                                     Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
+def get_open_filename(self: Any, title: str, filter_name: str, filter_mime: str, callback: Callable[[Gtk.FileDialog, Gio.AsyncResult], None]) -> None:
+    """Display a file open dialog."""
+    chooser = Gtk.FileDialog.new()
+    chooser.set_title(title)
+
     file_filter = Gtk.FileFilter()
     file_filter.set_name(filter_name)
     file_filter.add_mime_type(filter_mime)
-    chooser.add_filter(file_filter)
-    chooser.set_filter(file_filter)
-
-    response = chooser.run()
-    if response == Gtk.ResponseType.OK:
-        filename = chooser.get_filename()
-    chooser.destroy()
-    return filename
+    filters = Gio.ListStore.new(Gtk.FileFilter)
+    filters.append(file_filter)
+    chooser.set_filters(filters)
+    chooser.open(self.window, None, callback)
 
 
-def get_save_filename(self):
-    """Display a file save dialog and returns the filename."""
-    filename = None
-    chooser = Gtk.FileChooserDialog(_("Save Text File"), self.window,
-                                    Gtk.FileChooserAction.SAVE,
-                                    (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
-                                     Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
-    chooser.set_do_overwrite_confirmation(True)
-    response = chooser.run()
-    if response == Gtk.ResponseType.OK:
-        filename = chooser.get_filename()
-    chooser.destroy()
-    return filename
+def get_save_filename(self: Any) -> None:
+    """Display a file save dialog."""
+    chooser = Gtk.FileDialog.new()
+    chooser.set_title(_("Save Text File"))
+    chooser.save(self.window, None, self.save_dialog_callback)
 
 
-def error_message(self, message):
+def error_message(self: Any, message: str) -> None:
     """Display the string 'message' in an error dialog."""
-    dialog = Gtk.MessageDialog(self.window,
-                               Gtk.DialogFlags.MODAL |
-                               Gtk.DialogFlags.DESTROY_WITH_PARENT,
-                               Gtk.MessageType.ERROR, Gtk.ButtonsType.OK,
-                               message)
-    dialog.run()
-    dialog.destroy()
+    dialog = Gtk.AlertDialog.new(message)
+    dialog.show(self.window)
 
 
-def md5_of_file(fname):
+def md5_of_file(fname: str) -> str:
     """
     Calculate the md5 hash of the given file.
 
@@ -187,7 +173,7 @@ def md5_of_file(fname):
     return md5_hash.hexdigest()
 
 
-def get_data_file(root_dir, *path_segments):
+def get_data_file(root_dir: str, *path_segments: str) -> str:
     """Return the absolute path of the given data file."""
     # Where to look for data (ui and image files). By default,
     # this is ../data, relative your trunk layout
